#pragma once
#include <fc/filesystem.hpp>
#include <eosio/chain/block.hpp>
#include <eosio/chain/genesis_state.hpp>
#include <eosio/chain/block_log_config.hpp>
#include <future>

namespace eosio { namespace chain {

   namespace detail { class block_log_impl; }

   /* The block log is an external append only log of the blocks with a header. Blocks should only
    * be written to the log after they are irreversible as the log is append only. The log is a doubly
    * linked list of blocks. There is a secondary index file of only block positions that enables
    * O(1) random access lookup by block number.
    *
    * +---------------+----------------+---------------+----------------+-----+------------------+-------------------+
    * | Block 1 Entry | Pos of Block 1 | Block 2 Entry | Pos of Block 2 | ... | Head Block Entry | Pos of Head Block |
    * +---------------+----------------+---------------+----------------+-----+------------------+-------------------+
    *
    * +----------------+----------------+-----+-------------------+
    * | Pos of Block 1 | Pos of Block 2 | ... | Pos of Head Block |
    * +----------------+----------------+-----+-------------------+
    *
    * The block log can be walked in order by deserializing a block, skipping 8 bytes, deserializing a
    * block, repeat... The head block of the file can be found by seeking to the position contained
    * in the last 8 bytes the file. The block log can be read backwards by jumping back 8 bytes, following
    * the position, reading the block, jumping back 8 bytes, etc.
    *
    * Blocks can be accessed at random via block number through the index file. Seek to 8 * (block_num - 1)
    * to find the position of the block in the main file.
    *
    * The main file is the only file that needs to persist. The index file can be reconstructed during a
    * linear scan of the main file.
    */

   namespace bfs = boost::filesystem;

   class block_log {
      public:

         using config_type = block_log_config;

         block_log(const config_type& config);
         block_log(block_log&& other) = default;
         ~block_log();
         
         uint64_t append(const signed_block_ptr& block, packed_transaction::cf_compression_type segment_compression);

         // create futures for append, must call in order of blocks
         std::future<std::tuple<signed_block_ptr, std::vector<char>>>
            create_append_future(boost::asio::io_context& thread_pool,
                                 const signed_block_ptr& b, packed_transaction::cf_compression_type segment_compression);
         uint64_t append(std::future<std::tuple<signed_block_ptr, std::vector<char>>> f);

         void reset( const genesis_state& gs, const signed_block_ptr& genesis_block, packed_transaction::cf_compression_type segment_compression);
         void reset( const chain_id_type& chain_id, uint32_t first_block_num );
         
         block_id_type    read_block_id_by_num(uint32_t block_num)const;

         std::unique_ptr<signed_block>   read_signed_block_by_num(uint32_t block_num) const;

         const signed_block_ptr&        head() const;
         uint32_t                       first_block_num() const;

         static bool exists(const fc::path& data_dir);
         /**
          *  @param ids[in,out] The list of transaction ids to be pruned. After the member function returns,
          *                     it would be modified to contain the list of transaction ids that do not
          *                     exists in the specified block.
          *  @returns The number of transactions been pruned
          **/
         size_t prune_transactions(uint32_t block_num, vector<transaction_id_type>& ids);

         static const uint64_t npos = std::numeric_limits<uint64_t>::max();

         static const uint32_t min_supported_version;
         static const uint32_t max_supported_version;

         static fc::path repair_log( const fc::path& data_dir, uint32_t truncate_at_block = UINT32_MAX, const char* reversible_block_dir_name="" );

         static std::optional<genesis_state> extract_genesis_state( const fc::path& data_dir );

         static chain_id_type extract_chain_id( const fc::path& data_dir );

         static void construct_index(const fc::path& block_file_name, const fc::path& index_file_name);

         static bool contains_genesis_state(uint32_t version, uint32_t first_block_num);

         static bool contains_chain_id(uint32_t version, uint32_t first_block_num);

         static bool is_supported_version(uint32_t version);

         static bool trim_blocklog_front(const fc::path& block_dir, const fc::path& temp_dir, uint32_t truncate_at_block);
         static int  trim_blocklog_end(fc::path block_dir, uint32_t n);

         // used for unit test to generate older version blocklog
         static void set_version(uint32_t);
         uint32_t    version() const;
         uint32_t get_first_block_num() const;

         /**
          * @param n Only test 1 block out of every n blocks. If n is 0, it is maximum between 1 and the ceiling of the total number blocks divided by 8.
          */
         static void smoke_test(fc::path block_dir, uint32_t n);

         static void extract_blocklog(const fc::path& log_filename, const fc::path& index_filename,
                                      const fc::path& dest_dir, uint32_t start_block, uint32_t num_blocks);
         static void split_blocklog(const fc::path& block_dir, const fc::path& dest_dir, uint32_t stride);
         static void merge_blocklogs(const fc::path& block_dir, const fc::path& dest_dir);

       private:
         std::unique_ptr<detail::block_log_impl> my;
   };
} }
